# Copyright (C) 2019 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
"""
This module provides the CLI interface for conda-content-trust.
This is intended to provide a command-line signing and metadata update
interface.
"""
from argparse import ArgumentParser
from copy import deepcopy
from json import dumps

from . import __version__
from . import authentication as cct_authentication
from . import root_signing as cct_root_signing
from . import signing as cct_signing
from .common import (
    CCT_Error,
    PrivateKey,
    is_gpg_fingerprint,
    is_hex_key,
    load_metadata_from_file,
    write_metadata_to_file,
)


def cli(args=None):
    parser = ArgumentParser(
        description="Signing and verification tools for Conda",
        conflict_handler="resolve",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        help="Show the conda-content-trust version number and exit.",
        version="conda-content-trust %s" % __version__,
    )

    # Create separate parsers for the subcommands.
    sp = parser.add_subparsers(title="subcommands", dest="subcommand_name")

    # subcommand: sign-artifacts

    p_signrepo = sp.add_parser(
        "sign-artifacts",
        help=(
            "Given a repodata.json "
            "file, produce signatures over the metadata for each artifact listed, "
            "and update the repodata.json file with their individual signatures."
        ),
    )
    p_signrepo.add_argument(
        "repodata_fname",
        help=(
            "the filename of a repodata.json file from "
            "which to retrieve metadata for individual artifacts."
        ),
    )
    p_signrepo.add_argument(
        "private_key_fname",
        help=(
            "the filename of a file containing a "
            "hex string representation of an ed25519 private key to be used "
            "to sign each artifact's metadata"
        ),
    )

    # subcommand: verify-metadata

    p_verifymd = sp.add_parser(
        "verify-metadata",
        help=(
            "Uses the first (trusted) metadata file "
            "to verify the second (not yet trusted) metadata file.  For "
            "example, "
            '"conda-content-trust verify-metadata 4.root.json 5.root.json"'
            " to verify version 5 of root based on version 4 of root, or "
            '"conda-content-trust verify-metadata 4.root.json key_mgr.json" '
            "to verify key manager metadata based on version 4 of root."
        ),
    )
    p_verifymd.add_argument(
        "trusted_metadata_filename",
        help=(
            "the filename of the "
            "already-trusted metadata file that sets the rules for verifying "
            "the untrusted metadata file"
        ),
    )
    p_verifymd.add_argument(
        "untrusted_metadata_filename",
        help=("the filename of the " "(untrusted) metadata file to verify"),
    )

    # subcommand: modify-metadata

    p_modifymd = sp.add_parser(
        "modify-metadata",
        help=(
            "Interactive metadata modification.  Use "
            "this to produce a new version of a metadata file (like root.json "
            "or key_mgr.json), or correct an error in an unpublished metadata "
            "file, or review and sign a metadata file.  This increments "
            "version number / timestamp, reports changes on console, etc. For "
            'example, "conda-content-trust modify-metadata 8.root.json" '
            "for assistance in "
            "producing a new version of root (version 9) using version 8."
        ),
    )
    p_modifymd.add_argument(
        "metadata_filename",
        help=("the filename of the existing metadata " "file to modify"),
    )

    # If we're missing optional requirements for the next few options, note
    # that in their help strings.
    opt_reqs_str = ""
    if not cct_root_signing.SSLIB_AVAILABLE:
        opt_reqs_str = (
            "[Unavailable]: Requires optional "
            "dependencies: securesystemslib and gpg.  "
        )

    # subcommand: gpg-key-lookup
    p_gpglookup = sp.add_parser(
        "gpg-key-lookup",
        help=(
            opt_reqs_str
            + "Given the OpenPGP fingerprint of an ed25519-type OpenPGP key, fetch "
            "the actual ed25519 public key value of the underlying key."
        ),
    )
    p_gpglookup.add_argument(
        "gpg_key_fingerprint",
        help=(
            "the 40-hex-character key fingerprint (long keyid) for the "
            "OpenPGP/GPG key that you want to sign something with.  Do not "
            'add prefix "0x".'
        ),
    )

    # subcommand: gpg-sign

    p_gpgsign = sp.add_parser(
        "gpg-sign",
        help=(
            opt_reqs_str + "Sign a given "
            "piece of metadata using GPG instead of the usual signing "
            "mechanisms.  Takes an OpenPGP key fingerprint and a filename."
        ),
    )
    p_gpgsign.add_argument(
        "gpg_key_fingerprint",
        help=(
            "the 40-hex-character key fingerprint (long keyid) for the "
            "OpenPGP/GPG key that you want to sign something with.  Do not "
            'add prefix "0x".'
        ),
    )
    p_gpgsign.add_argument(
        "filename", help=("the filename of the file that will be signed")
    )

    args = parser.parse_args(args)

    if args.subcommand_name == "gpg-sign":
        # TODO: Validate arguments.

        # Strip any whitespace from the key fingerprint and lowercase it.
        # GPG pops out keys in a variety of whitespace arrangements and cases,
        # so this is necessary for convenience.
        gpg_key_fingerprint = "".join(args.gpg_key_fingerprint.split()).lower()

        cct_root_signing.sign_root_metadata_via_gpg(args.filename, gpg_key_fingerprint)

    elif args.subcommand_name == "sign-artifacts":
        with open(args.private_key_fname) as key_fobj:
            # Lower-case the hex string and ignore any whitespace before and
            # after it (in case someone adds some).
            private_key_hex = key_fobj.read().strip().lower()

        if not is_hex_key(private_key_hex):
            print(
                "ABORTED.  Expected key file to contain only a hex string "
                "representation of an ed25519 key.  It does not."
            )
            return

        cct_signing.sign_all_in_repodata(args.repodata_fname, args.private_key_hex)

    elif args.subcommand_name == "gpg-key-lookup":
        gpg_key_fingerprint = "".join(args.gpg_key_fingerprint.split()).lower()
        keyval = cct_root_signing.fetch_keyval_from_gpg(gpg_key_fingerprint)
        print("Underlying ed25519 public key value: " + str(keyval))

    elif args.subcommand_name == "modify-metadata":
        # `conda-content-trust update-metadata <metadata file to produce new version of>`

        # underlying functions: build_delegating_metadata,
        # load_metadata_from_file

        # given a metadata file, increment the version number and timestamps,
        # reporting the changes on the console

        # strip signatures

        # indicate what signatures are required

        # ask if the user wants to sign; query for the key hex or fname;
        # ideally, offer this functionality for both root and non-root keys.
        # For root metadata, we can (and should) also report which keys are
        # expected / still needed in order for the metadata to be verifiable
        # according to the old metadata and the new metadata

        old_metadata = load_metadata_from_file(args.metadata_filename)

        # new_metadata = cct_metadata_construction.interactive_modify_metadata(old_metadata)
        # if new_metadata is not None and new_metadata:
        #     write_metadata_to_file(new_metadata, args.metadata_filename)

        interactive_modify_metadata(old_metadata)

    elif args.subcommand_name == "verify-metadata":
        # `conda-content-trust verify-metadata <trusted delegating metadata> <untrusted
        # metadata> <(optional) role name>`

        # underlying functions: cct_authentication.verify_delegation,
        # load_metadata_from_file

        # takes two metadata files, the first being a trusted file that should
        # provide the verification criteria (expected keys and expected number
        # of keys) for the second file.  This should support root-root
        # verification (root chaining as currently implemented in
        # conda-content-trust) and delegation from one metadata type to another
        # (e.g. root to key_mgr)

        # conveys to the user whether or not the file is trusted, and for what
        # role.  e.g., would convey that the first file is (e.g.) a root
        # metadata file, that it provides a delegation to <role name>, and that
        # the <untrusted metadata> file provides <role name> and is signed
        # appropriately based on what the root metadata file requires of that
        # delegation.

        untrusted_metadata = load_metadata_from_file(args.untrusted_metadata_filename)

        trusted_metadata = load_metadata_from_file(args.trusted_metadata_filename)

        # TODO✅: Argument validation via the check_format_* calls.

        metadata_type = untrusted_metadata["signed"]["type"]

        if metadata_type == "root":
            # Verifying root has additional steps beyond verify_delegation.
            try:
                cct_authentication.verify_root(trusted_metadata, untrusted_metadata)
                print("Root metadata verification successful.")
                return 0  # success

            except CCT_Error as e:
                errorcode = 10
                errorstring = str(e)

        else:
            # Verifying anything other than root just uses verify_delegation
            # directly.
            try:
                cct_authentication.verify_delegation(
                    delegation_name=metadata_type,
                    untrusted_delegated_metadata=untrusted_metadata,
                    trusted_delegating_metadata=trusted_metadata,
                )
                print("Metadata verification successful.")
                return 0  # success

            except CCT_Error as e:
                errorcode = 20
                errorstring = str(e)

        # We should only get here if verification failed.
        print(
            "Verification of untrusted metadata failed.  Metadata "
            'type was "' + metadata_type + '".  Error reads:\n  "' + errorstring + '"'
        )
        return errorcode  # failure; exit code

    else:
        parser.print_help()


def interactive_modify_metadata(metadata):
    """ """

    # Update version if there is a version.
    # Update timestamp if there is a timestamp.
    #
    # Show metadata contents ('signed') -- pprint?
    #    indicate updated version/timestamp
    #
    # Changes phase:
    #    Prompt to
    #       (m) modify a value, (a) add a new entry, (d) delete an entry,
    #       (r) revert to original, (f) finish and sign ((move on to signing
    #       prompts))
    #
    # Signing phase:
    #   Show metadata again, ask if metadata looks right
    #   Show what keys the original was signed by and ask if those should be
    #     the keys used for the new version.
    #        ((Later: if root, vet against contents of new and old root versions))
    #   Prompt for key (raw key file, raw key data, or gpg key fingerprint)
    #   Sign using the given key (gpg if gpg, else normal signing mechanism).
    #   Write (making sure not to overwrite, and -- if root -- making sure to
    #     prepend "<version>." to root.json file.

    initial_metadata = metadata
    metadata = deepcopy(initial_metadata)

    try:
        import pygments
        import pygments.formatters
        import pygments.lexers
    except ImportError:
        print(
            "interactive modify-metadata mode employs pygments for syntax "
            "highlighting, if pygments is available.  pygments was not "
            "found, so the JSON contents will be... uglier than they "
            "would otherwise be.  If you would like syntax highlighting "
            "and prettier printing of JSON, you may install pygments."
        )
        pygments = None
        from pprint import pprint

    # Build the modification options and prompt.
    def promptfor(s):
        return input(F_INSTRUCT + "\n----- Please provide " + s + ENDC + ": ")

    def fn_write():
        fname = promptfor("a filename to save this metadata as")
        print("Writing to file....")
        write_metadata_to_file(metadata, fname)
        print("Modified metadata written!")
        return 1

    def fn_abort():
        # TODO✅: Ask to confirm.
        print(RED + BOLD + "\nAborting!\n" + ENDC)
        return 1

    def fn_addsig():
        if not cct_root_signing.SSLIB_AVAILABLE:
            print(
                F_OPTS + "Signing.  " + RED + "Please ABORT (control-c) if "
                "the metadata above is not EXACTLY what you want to sign!" + ENDC
            )
        key = promptfor(
            "a key: either:\n     - a 40-character-hex-string GPG PUBLIC "
            "key fingerprint\n"
            "       for GPG keys (e.g. root YubiKeys), or \n     - a "
            "64-character-hex-string PRIVATE key value for normal "
            "keys.\n\n     Whitespace will be removed and characters will "
            "be lowercased.\n     Key"
        )
        key = "".join(key.split()).lower()

        if is_hex_key(key):
            private_key = PrivateKey.from_hex(key)
            cct_signing.sign_signable(metadata, private_key)
            print(F_OPTS + "\n\n--- Successfully signed!  Please save." + ENDC)

        elif is_gpg_fingerprint(key):
            try:
                cct_root_signing.sign_root_metadata_dict_via_gpg(metadata, key)
            except:
                print(
                    F_OPTS
                    + "\n\n--- "
                    + RED
                    + "Signing FAILED."
                    + F_OPTS
                    + "  Do you have this key loaded in GPG on "
                    "this system?"
                )
            else:
                print(F_OPTS + "\n\n--- Successfully signed!  Please save." + ENDC)

        else:
            print(F_OPTS + RED + "Unable to recognize key.  Please try again." + ENDC)
        return 0

    def fn_remsig():
        return 0

    def fn_update():
        return 0

    def fn_adddel():
        return 0

    def fn_remdel():
        return 0

    def fn_thresh():
        delegation = promptfor(
            "a delegation name (one of the entries in the"
            '\n     "delegations" dictionary in the metadata above).  '
            "This will\n     be the delegation whose threshold number of "
            "required keys we\n     will change."
        )
        if delegation not in metadata["signed"]["delegations"]:
            print(
                F_OPTS + "\n\n--- " + RED + "Unable to find that delegation."
                "  Please try again." + ENDC
            )
            return 0

        new_thresh = promptfor(
            "a new threshold value.  The current value is "
            + str(metadata["signed"]["delegations"][delegation]["threshold"])
        )

        try:
            new_thresh = int(new_thresh)
            assert new_thresh >= 1
        except:
            print(
                F_OPTS + "\n--- " + RED + "Invalid value.  Expecting integer "
                "greater than or equal to 1.  Please try again." + ENDC
            )
            return 0

        metadata["signed"]["delegations"][delegation]["threshold"] = new_thresh

        print(F_OPTS + "\n--- Threshold successfully updated." + ENDC)

        return 0

    def fn_addkey():
        return 0

    def fn_remkey():
        return 0

    options = {
        0: [fn_write, "Done: write and save metadata"],
        1: [fn_abort, "Abort: discard changes -- abort without writing"],
        2: [fn_addsig, "Add a signature (sign with a key you have)"],
        3: [fn_remsig, "Remove a signature"],
        4: [fn_update, "Update any top-level dictionary entry"],
        5: [fn_adddel, "Add a delegation"],
        6: [fn_remdel, "Remove a delegation"],
        7: [fn_thresh, "Change the threshold number of keys for a delegation"],
        8: [fn_addkey, "Add an authorized key to a delegation"],
        9: [fn_remkey, "Remove an authorized key from a delegation"],
    }

    option_text = (
        F_INSTRUCT + "\n--- Please choose an operation by entering its "
        "number\n" + ENDC
    )
    for index in options:
        option_text += (
            "    "
            + F_LABEL
            + str(index)
            + ENDC
            + ": "
            + options[index][1]
            + ENDC
            + "\n"
        )

    done = False
    while not done:
        print(
            F_OPTS
            + BOLD
            + "\n\n---------------------\n--- Current metadata:\n---------------------\n"
            + ENDC
        )

        if pygments is not None:
            formatted_metadata = dumps(metadata, sort_keys=True, indent=4)
            print(
                pygments.highlight(
                    formatted_metadata.encode("utf-8"),
                    pygments.lexers.JsonLexer(),
                    pygments.formatters.TerminalFormatter(),
                )
            )
        else:
            pprint(metadata)

        print(option_text)
        selected = input(F_OPTS + "Choice: " + ENDC)
        try:
            selected = int(selected)
        except:
            print(RED + BOLD + "\nInvalid entry.  Try again.\n" + ENDC)
            continue
        if selected not in options:
            print(RED + BOLD + "\nInvalid entry.  Try again.\n" + ENDC)
            continue

        print(F_OPTS + '\nChose "' + options[selected][1] + '"' + ENDC)

        done = options[selected][0]()  # Run the func associated with the option.

    # Pull modified from debugging script
    # Pull modified from debugging script
    # Pull modified from debugging script


# Basic text formatting string constants
PINK = "\033[95m"
BLUE = "\033[94m"
CYAN = "\033[96m"
GREEN = "\033[92m"
YELLOW = "\033[93m"
RED = "\033[91m"
ENDC = "\033[0m"
BOLD = "\033[1m"
UNDERLINE = "\033[4m"

# Complete formats
F_LABEL = ENDC + UNDERLINE + BOLD + PINK
F_INSTRUCT = ENDC + BOLD + PINK
F_OPTS = ENDC + GREEN


if __name__ == "__main__":
    import sys

    exit_status = cli(sys.argv[1:])
    sys.exit(exit_status)



# Copyright (C) 2019 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
"""
This module contains functions that construct metadata and generate signing
keys.

Function Manifest for this Module

Key Creation:
  gen_keys
  gen_and_write_keys

Metadata Construction:
  build_delegating_metadata
  build_root_metadata         (wraps build_delegating_metadata)
"""
from datetime import timedelta

from .common import (
    SECURITY_METADATA_SPEC_VERSION,
    PrivateKey,
    PublicKey,
    checkformat_delegations,
    checkformat_natural_int,
    checkformat_string,
    checkformat_utc_isoformat,
    iso8601_time_plus_delta,
)

# Default expiration distance for repodata_verify.json.
REPODATA_VERIF_MD_EXPIRY_DISTANCE = timedelta(days=31)
ROOT_MD_EXPIRY_DISTANCE = timedelta(days=365)


def build_delegating_metadata(
    metadata_type, delegations=None, version=1, timestamp=None, expiration=None
):
    """
    # ✅ TODO: Docstring

    Builds delegating metadata, e.g. root.json, key_mgr.json.

    See metadata specification at:
    anaconda.atlassian.net/wiki/spaces/AD/pages/285147281/Conda+Security+Metadata+Specification

    Arguments:
        metadata_type:
            The type of this metadata (e.g. root or key_mgr).  This should
            match the intended filename (without .json)

        delegations (default {} )
            a dictionary defining the delegations this metadata makes.
            Each key is the role delegated to, with the value equal to a
            dictionary listing the acceptable public keys and threshold
            (number of signatures from distinct acceptable public keys) for the
            delegated role.  e.g.
            {   'root.json':
                    {'pubkeys': ['01'*32, '02'*32, '03'*32], 'threshold': 2},
                'key_mgr.json':
                    {'pubkeys': ['04'*32], 'threshold': 1}}

            If not provided, an empty dictionary (no delegations) will be used.

        version (default 1)
            the version of the metadata; root metadata must advance one version
            at a time (root chaining).  For other types of metadata, versions
            are advisory.

        timestamp (default: current system time)
            UTC time associated with the production of this metadata, in
            ISO8601 format (e.g. '2020-10-31T14:45:19Z')

        expiration (default: current system time plus ROOT_MD_EXPIRY_DISTANCE)
            UTC time beyond which this metadata should be considered expired
            and not verifiable by any client seeking new metadata
    """

    # Handle optional args
    if delegations is None:
        delegations = {}
    if timestamp is None:
        timestamp = iso8601_time_plus_delta(timedelta(0))  # now plus 0
    if expiration is None:
        expiration = iso8601_time_plus_delta(ROOT_MD_EXPIRY_DISTANCE)

    # Argument validation.  Note that this (checkformat_delegations) also
    # checks for duplicates in lists of keys, which is important to reduce the
    # odds of a developer introducing certain bugs that cause security issues
    # (multiple signatures from same key being treated as two unique sigs,
    # etc.)
    checkformat_string(metadata_type)
    # TODO: ✅⚠️ Consider a set of acceptable metadata types (root, key_mgr,
    #             channel_authority).  Have to be careful about backward
    #             compatibility, though....
    checkformat_utc_isoformat(timestamp)
    checkformat_utc_isoformat(expiration)
    checkformat_natural_int(version)
    checkformat_delegations(delegations)

    md = {
        "type": metadata_type,
        "version": version,
        "metadata_spec_version": SECURITY_METADATA_SPEC_VERSION,
        "timestamp": timestamp,
        "expiration": expiration,
        "delegations": delegations,
    }

    # # This very redundant, but might be useful as defensive code.
    # checkformat_delegating_metadata(wrap_as_signable(md)

    return md


def build_root_metadata(
    root_version,
    root_pubkeys,
    root_threshold,
    key_mgr_pubkeys,
    key_mgr_threshold,
    root_timestamp=None,
    root_expiration=None,
):
    """
    Wrapper for build_delegating_metadata().  Helpfully requires root to list
    itself and key_mgr in its delegations.

    # ✅ TODO: Docstring

    # ✅ TODO: Expand build_root_metadata flexibility for
    #          directly-root-delegated roles (i.e. in addition to channeler).
    """

    # Note that argument validation is performed in the
    # build_delegation_metadata call below.  So is some of the optional
    # argument default setting (timestamp).  We set expiration explicitly here
    # in case the defaults for generic delegating metadata and root metadata
    # diverge later.
    # Note that it is probably best to provide less revealing timestamps for
    # root metadata generation (00:00:00 of a past day), since it is a manual
    # process and patterns in that information might be somewhat useful to a
    # sophisticated attacker.
    if root_expiration is None:
        root_expiration = iso8601_time_plus_delta(ROOT_MD_EXPIRY_DISTANCE)
    # if channeler_pubkeys is None:
    #     channeler_pubkeys = []
    # if channeler_threshold = None:
    #     channeler_threshold = max(1, len(channeler_pubkeys))

    delegations = {
        "root": {"pubkeys": root_pubkeys, "threshold": root_threshold},
        "key_mgr": {"pubkeys": key_mgr_pubkeys, "threshold": key_mgr_threshold},
    }

    root_md = build_delegating_metadata(
        metadata_type="root",
        delegations=delegations,
        version=root_version,
        timestamp=root_timestamp,
        expiration=root_expiration,
    )

    return root_md


def gen_and_write_keys(fname):
    """
    Generate an ed25519 key pair, then write the key files to disk.

    Given fname, write the private key to fname.pri, and the public key to
    fname.pub. Performs no filename validation, etc.  Also returns the private
    key object and the public key object, in that order.
    """

    # Create an ed25519 key pair, employing OS random generation.
    # Note that this just has the private key sitting around.  In the real
    # implementation, we'll want to use an HSM equipped with an ed25519 key.
    private, public = gen_keys()

    # Write the actual bytes of the key values to disk as requested.
    # Note that where the private key is concerned, we're just grabbing the
    # not-encrypted private key value.
    with open(fname + ".pri", "wb") as fobj:
        fobj.write(PrivateKey.to_bytes(private))
    with open(fname + ".pub", "wb") as fobj:
        fobj.write(PublicKey.to_bytes(public))

    return private, public


def gen_keys():
    """
    Generate an ed25519 key pair and return it (private key, public key).

    Returns two objects:
      - a conda_content_trust.common.PrivateKey, a subclass of
        cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PrivateKey
      - a conda_content_trust.common.PublicKey, a subclass of
        cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PublicKey
    """
    # Create an ed25519 key pair, employing OS random generation.
    # Note that this just has the private key sitting around.  In the real
    # implementation, we'll want to use an HSM equipped with an ed25519 key.
    private = PrivateKey.generate()
    public = private.public_key()

    return private, public


# This function is not in use.  It's here for reference, in case it's useful
# again in the future.
# def build_repodata_verification_metadata(
#         repodata_hashmap, channel=None, expiry=None, timestamp=None):
#     """
#     # TODO: ✅ Full docstring.

#     # TODO: ✅ Contemplate the addition of "version" to this metadata.  As yet,
#     #          the timestamp serves our purposes....

#     Note that if expiry or timestamp are not provided or left as None, now is
#     used for the timestamp, and expiry is produced using a default expiration
#     distance, via iso8601_time_plus_delta().  (It does not mean no expiration!)

#     Channel may be optionally specified, and is only included if specified.

#     Sample input (repodata_hashmap):
#     {
#         "noarch/current_repodata.json": "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
#         "noarch/repodata.json": "...",
#         "noarch/repodata_from_packages.json": "...",
#         "osx-64/current_repodata.json": "...",
#         "osx-64/repodata.json": "...",
#         "osx-64/repodata_from_packages.json": "..."
#     }

#     Sample output:
#         See metadata specification (version defined by
#         SECURITY_METADATA_SPEC_VERSION) for definition and samples of type
#         "Repodata Verification Metadata".
#     """

#     if expiry is None:
#         expiry = iso8601_time_plus_delta(REPODATA_VERIF_MD_EXPIRY_DISTANCE)

#     if timestamp is None:
#         timestamp = iso8601_time_plus_delta(timedelta(0))

#     # TODO: ✅ More argument validation: channel,
#     checkformat_utc_isoformat(expiry)
#     checkformat_utc_isoformat(timestamp)
#     if not ( # dict with string keys and 32-byte-hash-as-hex-string values
#             isinstance(repodata_hashmap, dict)
#             and all([isinstance(x, string_types) for x in repodata_hashmap])
#             and all([is_hex_hash(repodata_hashmap[x]) for x in repodata_hashmap])):
#         raise ValueError(
#                 'Argument repodata_hashmap must be a dictionary with strings '
#                 'as keys (filenames of repodata.json files), and values that '
#                 'are 64-character hex strings representing 32-byte hashes (of '
#                 'those repodata.json files)')

#     # TODO: ✅ Really have to make TypeError and ValueError usages consistent
#     #       with norms throughout this codebase.


#     rd_v_md = {
#             'type': 'repodata_verify',
#             # (Take advantage of iso8601_time_plus_delta() to get current time
#             #  in the ISO8601 UTC format we want.)
#             'timestamp': timestamp, # version->timestamp in spec v 0.0.5
#             'metadata_spec_version': SECURITY_METADATA_SPEC_VERSION,
#             'expiration': expiry,
#             'secured_files': repodata_hashmap}

#     if channel is not None:
#         rd_v_md['channel'] = channel

#     return rd_v_md



print("import: 'conda_content_trust'")
import conda_content_trust

print("import: 'conda_content_trust.authentication'")
import conda_content_trust.authentication

print("import: 'conda_content_trust.common'")
import conda_content_trust.common

print("import: 'conda_content_trust.root_signing'")
import conda_content_trust.root_signing

print("import: 'conda_content_trust.signing'")
import conda_content_trust.signing




# Copyright (C) 2019 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
"""
This module contains functions that provide format validation, serialization,
and some key transformations for the pyca/cryptography library.  These are used
across conda_content_trust modules.

Function Manifest for this Module, by Category

Encoding:
  x  canonserialize

Formats and Validation:
     PrivateKey  -- extends cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PrivateKey
     PublicKey   -- extends cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PublicKey
     checkformat_string
  x  is_hex_string
  x  is_hex_signature
  r  is_hex_key
  r  checkformat_hex_key
  r  checkformat_list_of_hex_keys
  x  is_signable
  x  checkformat_byteslike
  x  checkformat_natural_int
  x  checkformat_expiration_distance
  x  checkformat_utc_isoformat
  x  checkformat_gpg_fingerprint
     is_gpg_fingerprint
  x  checkformat_gpg_signature
     is_gpg_signature
     checkformat_any_signature
     checkformat_delegation
     checkformat_delegations
     checkformat_delegating_metadata
  x  iso8601_time_plus_delta

Crypto Utility:
   x sha512256
   x keyfiles_to_keys
   x keyfiles_to_bytes

Exceptions:
    CCT_Error
        SignatureError
        MetadataVerificationError
        UnknownRoleError
"""
from __future__ import annotations

from binascii import hexlify, unhexlify
from datetime import datetime, timedelta
from json import dumps, load
from typing import Any, Protocol

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519

# specification version for the metadata produced by conda-content-trust
# Details in the Conda Security Metadata Specification.  Note that this
# version string is parsed via setuptools's packaging.version library, and so
# supports PEP 440; however, we should use a limited subset that is numerical
# only, and according to SemVer principles.
# PEP 440 compatibility:
#   > None is not re.match(r'^([1-9]\d*!)?(0|[1-9]\d*)(\.(0|[1-9]\d*))*((a|b|rc)(0|[1-9]\d*))?(\.post(0|[1-9]\d*))?(\.dev(0|[1-9]\d*))?$', version_string)
# SemVer compatibility:
#   > None is not re.match(r'^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$', version_string)
# Try, however, to keep to three simple numeric elements separated by periods,
# i.e., things that match this subset of SemVer:
#   > None is not re.match(r'^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)$', version_string)
SECURITY_METADATA_SPEC_VERSION = "0.6.0"

# The only types we're allowed to wrap as "signables" and sign are
# the JSON-serializable types.  (There are further constraints to what is
# JSON-serializable in addition to these type constraints.)
SUPPORTED_SERIALIZABLE_TYPES = [dict, list, tuple, str, int, float, bool, type(None)]

# These are the permissible strings in the "type" field of delegating metadata.
SUPPORTED_DELEGATING_METADATA_TYPES = ["root", "key_mgr"]  # May be loosened later.


class CCT_Error(Exception):
    """
    All errors we raise that are not ValueErrors, TypeErrors, or
    certain errors from securesystemslib should be instances of this class or
    of subclasses of this class.
    """


class SignatureError(CCT_Error):
    """
    Indicates that a signable cannot be verified due to issues with the
    signature(s) inside it.
    """


class MetadataVerificationError(CCT_Error):
    """
    Indicates that a chain of authority metadata cannot be verified (e.g.
    a metadata update is found on the repository, but could not be
    authenticated).
    """


class UnknownRoleError(CCT_Error):
    """
    Indicates that a piece of role metadata (like root.json, or key_mgr.json)
    was expected but not found.
    """


def canonserialize(obj):
    """
    Given a JSON-compatible object, does the following:
     - serializes the dictionary as utf-8-encoded JSON, lazy-canonicalized
       such that any dictionary keys in any dictionaries inside <dictionary>
       are sorted and indentation is used and set to 2 spaces (using json lib)

    TODO: ✅ Implement the serialization checks from serialization document.

    Note that if the provided object includes a dictionary that is *indexed*
    by both strings and integers, a TypeError will be raised complaining about
    comparing strings and integers during the sort.  (Each dictionary in an
    object must be indexed only by strings or only by integers.)
    """

    # Try converting to a JSON string.
    try:
        # TODO: In the future, assess whether or not to employ more typical
        #       practice of using no whitespace (instead of NLs and 2-indent).
        json_string = dumps(obj, indent=2, sort_keys=True)
    except TypeError:
        # TODO: ✅ Log or craft/use an appropriate exception class.
        raise

    return json_string.encode("utf-8")


def load_metadata_from_file(fname):
    # TODO ✅: Argument validation for fname.  Consider adding "pathvalidate"
    #          as a dependency, and calling its sanitize_filename() here.

    with open(fname, "rb") as fobj:
        metadata = load(fobj)

    # TODO ✅: Consider validating what is read here, for everywhere.

    return metadata


def write_metadata_to_file(metadata, filename):
    """
    Canonicalizes and serializes JSON-friendly metadata, and writes that to the
    given filename.
    """

    # TODO ✅: Argument validation for filename.  Consider adding
    #          "pathvalidate" as a dependency, and calling its
    #          sanitize_filename() here.

    metadata = canonserialize(metadata)

    with open(filename, "wb") as fobj:
        fobj.write(metadata)


class MixinKey:
    """
    This is a mix-in (https://www.ianlewis.org/en/mixins-and-python) for the
    PrivateKey and PublicKey classes, specifically.  It provides some
    convenience functions.
    """

    @classmethod
    def to_hex(cls, key):
        return hexlify(cls.to_bytes(key)).decode("utf-8")

    @classmethod
    def is_equivalent_to(cls, k1, k2):
        """
        Given Ed25519PrivateKey or Ed25519PublicKey objects, determines if the
        underlying key data is identical.
        """
        checkformat_key(k2)
        if type(k1) is not type(k2):
            return False
        return cls.to_bytes(k1) == cls.to_bytes(k2)

    @classmethod  # a class method for inheritors of this mix-in
    def from_hex(cls, key_value_in_hex):
        # from_private_bytes() and from_public_bytes() both check length (32),
        # but do not produce helpful errors if the argument provided it is not
        # the right type, so we'll do that here before calling them.
        checkformat_hex_key(key_value_in_hex)
        key_value_in_bytes = unhexlify(key_value_in_hex)
        new_object = cls.from_bytes(key_value_in_bytes)
        checkformat_key(new_object)
        return new_object


class PrivateKey(MixinKey, ed25519.Ed25519PrivateKey):
    """
    This class expands the class
    cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PrivateKey
    very slightly, adding some functionality from MixinKey.

    Note on the sign() method:
        We preserve Ed25519PrivateKey's sign method unchanged.  The sign()
        method is deterministic and does not depend at any point on the ability
        to generate random data (unlike the key generation).  The returned
        value for sign() is a length 64 bytes() object, a raw ed25519
        signature.
    """

    @classmethod
    def to_bytes(cls, key):
        return key.private_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PrivateFormat.Raw,
            encryption_algorithm=serialization.NoEncryption(),
        )

    @classmethod
    def from_bytes(cls, key_value_in_bytes):
        """
        Constructs an object of the class based on the given key value.
        The "cryptography" library provides from_public_bytes() and
        from_private_bytes() class methods for Ed25519PublicKey and
        Ed25519PrivateKey classes in place of constructors.  We extend provide
        a single API for those, and make the created objects objects of the
        subclass using this mix-in.
        """
        # from_private_bytes() and from_public_bytes() both check length (32),
        # but do not produce helpful errors if the argument provided it is not
        # the right type, so we'll do that here before calling them.
        checkformat_byteslike(key_value_in_bytes)
        return super().from_private_bytes(key_value_in_bytes)


class PublicKey(MixinKey, ed25519.Ed25519PublicKey):
    """
    This class expands the class
    cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PublicKey
    very slightly, adding some functionality from MixinKey.

    We preserve Ed25519PublicKey's verify() method unchanged.
    """

    @classmethod
    def to_bytes(cls, key):
        """
        Pops out the nice, tidy bytes of a given ed25519 key object, public or
        private.
        """
        return key.public_bytes(
            serialization.Encoding.Raw, serialization.PublicFormat.Raw
        )

    @classmethod
    def from_bytes(cls, key_value_in_bytes):
        """
        Constructs an object of the class based on the given key value.
        The "cryptography" library provides from_public_bytes() and
        from_private_bytes() class methods for Ed25519PublicKey and
        Ed25519PrivateKey classes in place of constructors.  We extend provide
        a single API for those, and make the created objects objects of the
        subclass using this mix-in.
        """
        # from_private_bytes() and from_public_bytes() both check length (32),
        # but do not produce helpful errors if the argument provided it is not
        # the right type, so we'll do that here before calling them.
        checkformat_byteslike(key_value_in_bytes)
        return super().from_public_bytes(key_value_in_bytes)


# No....  For now, I'll stick with the raw dictionary representations.
# If function profusion makes it inconvenient for folks to use this library,
# it MAY then be time to make signatures into class objects... but it's
# probably best to avoid that potential complexity and confusion.
# class Signature():
#     def __init__(self, ):
#         self.is_gpg_sig = False


# ✅ TODO: Consider a schema definitions module, e.g. PyPI project "schema"
def is_hex_string(hex_string: Any) -> bool:
    """
    Returns True if hex is a hex string with no uppercase characters, no spaces,
    etc.  Else, False.
    """
    try:
        checkformat_hex_string(hex_string)
        return True
    except (ValueError, TypeError):
        return False


HexString = str


def checkformat_hex_string(hex_string: Any) -> HexString:
    """
    Throws TypeError if s is not a string.
    Throws ValueError if the given string is not a string of hexadecimal
    characters (upper-case not allowed to prevent redundancy).
    """
    bytes.fromhex(hex_string)
    # isalnum() checks for no whitespace which bytes.fromhex() would allow.
    if not hex_string.isalnum() or hex_string.lower() != hex_string:
        raise ValueError(
            "Expected a hex string; non-hexadecimal or upper-case character found."
        )

    return hex_string


def is_hex_signature(hex_signature: Any) -> bool:
    """
    Returns True if key is a hex string with no uppercase characters, no
    spaces, no '0x' prefix(es), etc., and is 128 hexadecimal characters (the
    correct length for an ed25519 signature, 64 bytes of raw data represented
    as 128 hexadecimal characters).
    Else, returns False.
    """
    if is_hex_string(hex_signature) and len(hex_signature) == 128:
        return True

    return False


def is_hex_key(hex_key: Any) -> bool:
    """
    Returns True if key is a hex string with no uppercase characters, no
    spaces, no '0x' prefix(es), etc., and is 64 hexadecimal characters (the
    correct length for an ed25519 key, 32 bytes of raw data represented as 64
    hexadecimal characters).
    Else, returns False.
    """
    try:
        checkformat_hex_key(hex_key)
        return True
    except (TypeError, ValueError):
        return False


def is_signable(signable: Any) -> bool:
    """
    Returns True if the given dictionary is a signable dictionary as produced
    by wrap_as_signable.  Note that there MUST be no additional elements beyond
    'signed' and 'signable' in the dictionary.  (The only data in the envelope
    outside the signed portion of the data should be the signatures; what's
    outside of 'signed' is under attacker control.)
    """
    return (
        isinstance(signable, dict)
        and set(signable) == {"signatures", "signed"}
        and isinstance(signable["signatures"], dict)
        and type(signable["signed"]) in SUPPORTED_SERIALIZABLE_TYPES
    )


Signable = dict


# TODO: ✅ Consolidate: switch to use of this wherever is_a_signable is called
#          and then an error is raised if the result is False.
def checkformat_signable(signable: Any) -> Signable:
    if not is_signable(signable):
        raise TypeError(
            "Expected a signable dictionary, but the given argument "
            "does not match expectations for a signable dictionary "
            '(must be a dictionary containing only keys "signatures" and '
            '"signed", where the value for key "signatures" is a dict '
            'and the value for key "signed" is a supported serializable '
            "type (" + str(SUPPORTED_SERIALIZABLE_TYPES) + ")"
        )

    return signable


class BytesLike(Protocol):
    def decode(self, *args, **kwargs) -> str:
        ...


def checkformat_byteslike(byteslike: Any) -> BytesLike:
    if not hasattr(byteslike, "decode"):
        raise TypeError("Expected a bytes-like object with a decode method.")

    return byteslike


def checkformat_natural_int(natural_int: Any) -> int:  # Annotated[int, ">= 1"]
    # Technically a TypeError or ValueError, depending, but meh.
    if int(natural_int) != natural_int or natural_int < 1:
        raise ValueError("Expected an integer >= 1.")

    return natural_int


# This is not yet widely used.
# TODO: ✅ See to it that anywhere we're checking for a string, we use this.
def checkformat_string(string: Any) -> str:
    if not isinstance(string, str):
        raise TypeError("Expecting a string")

    return string


def checkformat_expiration_distance(expiration_distance: Any) -> timedelta:
    if not isinstance(expiration_distance, timedelta):
        raise TypeError(
            "Expiration distance must be a datetime.timedelta object. "
            "Instead received a " + str(type(expiration_distance))
        )

    return expiration_distance


HexKey = HexString  # Annotated[HexString, "len() == 64"]


def checkformat_hex_key(hex_key: Any) -> HexKey:
    checkformat_hex_string(hex_key)

    if 64 != len(hex_key):
        raise ValueError("Expected a 64-character hex string representing a key value.")

    return hex_key


def checkformat_list_of_hex_keys(list_of_hex_keys: Any) -> list[HexKey]:
    """
    Note that this rejects any list of keys that includes any exact duplicates.
    """
    if not isinstance(list_of_hex_keys, list):
        raise TypeError(
            "Expected a list of 64-character hex strings representing keys."
        )

    for hex_key in list_of_hex_keys:
        checkformat_hex_key(hex_key)

    if len(set(list_of_hex_keys)) != len(list_of_hex_keys):
        raise ValueError(
            "The given list of keys in hex string form contains duplicates.  "
            "Duplicates are not permitted."
        )

    return list_of_hex_keys


def checkformat_utc_isoformat(date_string: Any) -> str:
    try:
        datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%SZ")
    except ValueError:
        # ValueError: date_string does not match format '%Y-%m-%dT%H:%M:%SZ'
        raise TypeError(
            "The provided string appears not to be a datetime string "
            "formatted as an ISO8601 UTC-specific datetime (e.g. "
            '"1999-12-31T23:59:59Z").'
        ) from None

    return date_string


def is_gpg_fingerprint(gpg_fingerprint: Any) -> bool:
    """
    True if the given value is a hex string of length 40 (representing a
    20-byte SHA-1 value, which is what OpenPGP/GPG uses as a key fingerprint).
    """
    try:
        checkformat_gpg_fingerprint(gpg_fingerprint)
        return True
    except (TypeError, ValueError):
        return False


GPGFingerprint = HexKey  # Annotated[HexKey, "len()==40"]


def checkformat_gpg_fingerprint(gpg_fingerprint: Any) -> GPGFingerprint:
    """
    See is_gpg_fingerprint.  Raises a TypeError if is_gpg_fingerprint is not
    True.
    """
    if len(gpg_fingerprint) != 40:
        raise ValueError(
            'The given value, "' + str(gpg_fingerprint) + '", is not a full '
            "GPG fingerprint (40 hex characters)."
        )

    # ⚠️ Yes, the following is a redundant test.  Please leave it here in case
    #    code changes elsewhere.
    # Prevent multiple possible representations of keys.  There are security
    # implications.  For example, we cannot permit two signatures from the
    # same key -- with the key represented differently -- to count as two
    # signatures from distinct keys.
    # local hex test. isalnum() checks for no whitespace.
    bytes.fromhex(gpg_fingerprint)
    if not gpg_fingerprint.isalnum() or gpg_fingerprint.lower() != gpg_fingerprint:
        raise ValueError(
            "Expected a hex string; non-hexadecimal or upper-case character found."
        )

    return gpg_fingerprint


def is_gpg_signature(gpg_signature: Any) -> bool:
    # TODO: ✅ docstring based on docstring from checkformat_gpg_signature

    try:
        checkformat_gpg_signature(gpg_signature)
        return True
    except (ValueError, TypeError):
        return False


GPGSignature = dict


def checkformat_gpg_signature(gpg_signature: Any) -> GPGSignature:
    """
    Raises a TypeError if the given object is not a dictionary representing a
    signature in a format that we expect.

    This is similar to BUT NOT THE SAME AS that produced by
    securesystemslib.gpg.functions.create_signature(), conforming to
    securesystemslib.formats.GPG_SIGNATURE_SCHEMA.

    We use a slightly different format in order to include the raw ed25519
    public key value. This is the format we expect for Root signatures.

    If the given object matches the format, returns silently.
    """
    if not isinstance(gpg_signature, dict):
        raise TypeError(
            "OpenPGP signatures objects must be dictionaries.  Received "
            "type " + str(type(gpg_signature)) + " instead."
        )

    if sorted(list(gpg_signature.keys())) not in [
        ["other_headers", "signature"],
        ["other_headers", "see_also", "signature"],
    ]:
        raise ValueError(
            'OpenPGP signature objects must include a "signature" and an '
            '"other_headers" entry, and may include a "see_also" entry.  No '
            "other entries are permitted."
        )

    if not is_hex_string(gpg_signature["other_headers"]):
        raise ValueError(
            '"other_headers" entry in OpenPGP signature object must be a ' "hex string."
        )
        # TODO ✅: Determine if we can constrain "other_headers" beyond
        #          limiting it to a hex string.  (No length constraint is
        #          provided here, for example.)

    if not is_hex_signature(gpg_signature["signature"]):
        raise ValueError(
            '"signature" entry in OpenPGP signature obj must be a hex '
            "string representing an ed25519 signature, 128 hex characters "
            "representing 64 bytes of data."
        )

    if "see_also" in gpg_signature:
        checkformat_gpg_fingerprint(gpg_signature["see_also"])

    return gpg_signature


def is_signature(signature: Any) -> bool:
    """
    Returns True if signature_obj is a dictionary representing an ed25519
    signature, either in the conda-content-trust normal format, or
    the format for a GPG signature.

    See conda_content_trust.common.checkformat_signature() docstring for more details.
    """
    try:
        checkformat_signature(signature)
        return True
    except (TypeError, ValueError):
        return False


Signature = dict


def checkformat_signature(signature: Any) -> Signature:
    """
    Raises a TypeError if the given object is not a dictionary.
    Raises a ValueError if the given object is a dictionary, but is not in
    our generalized signature format (supports both raw ed25519 signatures
    OpenPGP/GPG signatures).

    If the given object matches the format, returns silently.

    The generalized signature format is:
    {
     (REQUIRED)      'signature': <64-byte value ed25519 signature, as 128 hex chars>,
     (GPG SIGS ONLY) 'other_headers': <hex string of irrelevant OpenPGP data hashed in the signature digest>,
     (OPTIONAL)      'see_also': <40-hex-character SHA1 OpenPGP/GPG key identifier, for diagnostic purposes>
    }
    Examples:
        { 'signature': 'deadbeef'*32}      # normal ed25519 signature (no OpenPGP)

        { 'signature': 'deadbeef'*32,      # OpenPGP ed25519 signature
          'other_headers': 'deadbeef'*??}  # extra info OpenPGP insists on signing over

        { 'signature': 'deadbeef'*32,      # OpenPGP ed25519 signature
          'other_headers': 'deadbeef'*??,
          'see_also': 'deadbeef'*10}}      # listing an OpenPGP key fingerprint
    """
    if not isinstance(signature, dict):
        raise TypeError("Expected a signature object, of type dict.")
    elif not ("signature" in signature and is_hex_signature(signature["signature"])):
        # Even the minimal required element is not correct, so...
        raise ValueError(
            "Expected a dictionary representing an ed25519 signature as a "
            "128-character hex string.  This requires at least key "
            '"signature", with value a 128-character hexadecimal string '
            "representing a (64-byte) ed25519 signature."
        )

    # simple ed25519 signature, not an OpenPGP signature
    elif len(signature) == 1:
        # If this is a simple ed25519 signature, and not an OpenPGP/GPG
        # signature, then we're all set, since 'signature' is included and
        # has a reasonable value.
        return signature

    # Permit an OpenPGP (GPG / RFC 4880) signature noted as defined in
    # function is_gpg_signature.
    elif is_gpg_signature(signature):
        return signature

    else:
        raise ValueError(
            "Provided signature does not have the correct format for a "
            "signature object (neither simple ed25519 sig nor OpenPGP "
            "ed25519 sig)."
        )


Delegation = dict


def checkformat_delegation(delegation: Any) -> Delegation:
    """
    A dictionary specifying public key values and threshold of keys
    e.g.
        {   'pubkeys': ['ff'*32, '1e'*32],
            'threshold': 1}
    threshold must be an integer >= 1. pubkeys must be a list of hexadecimal
    representations of ed25519 public keys.

    Note that because drafts are allowed, we do not demand here that the list
    of pubkeys include enough keys to meet threshold.  Not listing pubkeys yet
    is okay during writing, but when verifying metadata, one should not accept
    delegations with impossible-to-meet requirements (len(pubkeys) < threshold)
    """
    if not isinstance(delegation, dict):
        raise TypeError(
            "Delegation information must be a dictionary specifying "
            '"pubkeys" and "threshold" elements.'
        )
    elif not (
        set(delegation) == {"threshold", "pubkeys"}
        and delegation["threshold"] >= 1
        and isinstance(delegation["pubkeys"], list)
        and all([is_hex_key(k) for k in delegation["pubkeys"]])
    ):
        raise ValueError(
            "Delegation information must be a dictionary specifying "
            'exactly two elements: "pubkeys" (assigned a list of '
            "64-character hex strings representing individual ed25519 "
            'public keys) and "threshold", assigned an integer >= 1.'
        )

    # We have the right type, and the right keys.  Check the values.
    checkformat_list_of_hex_keys(delegation["pubkeys"])
    checkformat_natural_int(delegation["threshold"])

    return delegation


def checkformat_delegations(delegations: Any) -> dict[str, Delegation]:
    """
    A dictionary specifying a delegation for any number of role names.
    Index: rolename.  Value: delegation (see checkformat_delegation).
    e.g.
        {   'root.json':
                {'pubkeys': ['01'*32, '02'*32, '03'*32], # each is a lower-case hex string w/ length 64
                 'threshold': 2},                        # integer >= 1
            'channeler.json':
                {'pubkeys': ['04'*32], 'threshold': 1}}
    """
    if not isinstance(delegations, dict):
        raise TypeError(
            '"Delegations" information must be a dictionary indexed by '
            "role names, with values equal to dictionaries that each "
            'specify elements "pubkeys" and "threshold".'
        )

    for index in delegations:
        checkformat_string(index)
        checkformat_delegation(delegations[index])

    return delegations


DelegatingMetadata = Signable


def checkformat_delegating_metadata(metadata: Any) -> DelegatingMetadata:
    """
    Validates argument "metadata" as delegating metadata.  Passes if it is,
    raises a TypeError or ValueError if it is not.

    The required format is a dictionary containing all contents of a delegating
    metadata file, like root.json or key_mgr.json.  (This includes both the
    signatures portion and the signed contents portion, in the usual envelope
    -- see also checkformat_signable.)

    The required structure:
        {
          'signatures': {}, # for each entry in the 'signatures' dict:
                            #  - the key must pass checkformat_hex_key
                            #  - the value must pass checkformat_signature()
                            #    or checkformat_gpg_signature()
          'signed': {
            'type': <type>: # must match a string in SUPPORTED_DELEGATING_METADATA_TYPES (e.g. 'root')
            'metadata_spec_version': <SemVer string>,
            'delegations': {}, # value must pass checkformat_delegations()
            'expiration': <date>, # date must pass checkformat_utc_isoformat()

            # The 'signed' dict must always include either a 'timestamp' entry,
            # a 'version' entry, or both.  Further, in root metadata the
            # 'signed' dict must always include a 'version' entry (to support
            # root chaining).

            'timestamp': <date>, # if included, must pass checkformat_utc_isoformat()
            'version': <integer> # if included, must pass checkformat_natural_int(), i.e. be an integer >= 1
          }
        }

    e.g.
        {
          "signatures": {       # 0 or more signatures over the signed contents
            <public key>: {
              "other_headers": "04001608001d162104f075dd2f6f4cb3bd76134bbb81b6ca16ef9cd58905025f0bf546",
              "signature": "ab3e03385f757da74e08b46f1bf82709fbc2ce21823c28e2f0e3452415e2a9f1e2c82e418cc44e2908618cf0c7375f32fe0a5a75494909a59a82875ebc703c02",
            },
            ...
          },
          "signed": {           # signed contents
            "delegations": {
              "key_mgr.json": {
                "pubkeys": [
                  "013ddd714962866d12ba5bae273f14d48c89cf0773dee2dbf6d4561e521c83f7"
                ],
                "threshold": 1
              },
              "root.json": {
                "pubkeys": [
                  "bfbeb6554fca9558da7aa05c5e9952b7a1aa3995dede93f3bb89f0abecc7dc07"
                ],
                "threshold": 1
              }
            },
            "expiration": "2021-07-13T05:46:45Z",
            "metadata_spec_version": "0.1.0",
            "timestamp": "2020-07-13T05:46:45Z",
            "type": "root",
            "version": 1
          }
        }
    """
    # Signing envelope required
    checkformat_signable(metadata)

    for k in metadata["signatures"]:
        checkformat_any_signature(metadata["signatures"][k])

    contents = metadata["signed"]

    for entry in [  # required fields
        "type",
        "metadata_spec_version",
        "delegations",
        "expiration",
    ]:
        if entry not in contents:
            raise ValueError(
                'Expected a "' + str(entry) + '" entry in the given '
                "delegating metadata."
            )

    checkformat_string(contents["type"])
    if contents["type"] not in SUPPORTED_DELEGATING_METADATA_TYPES:
        raise ValueError(
            'Given type entry ("' + contents["type"] + '") is not '
            "one of the supported types of delegating metadata."
        )

    checkformat_string(contents["metadata_spec_version"])
    # TODO ✅⚠️: For metadata_spec_version, add semantic versioning checks:
    #                 - check format
    #                 - check for compatibility with common.SECURITY_METADATA_SPEC_VERSION

    checkformat_delegations(contents["delegations"])

    checkformat_utc_isoformat(contents["expiration"])

    # Timestamp and/or Version:
    if "timestamp" not in contents and "version" not in contents:
        raise ValueError(
            'All metadata must include a "version" entry, or a '
            '"timestamp" entry, or both.'
        )

    if contents["type"] == "root" and "version" not in contents:
        raise ValueError("Root metadata must specify its version number.")
    # Catch a possible future coding error at the PR stage, here where the
    # assumption is being made.
    assert "root" in SUPPORTED_DELEGATING_METADATA_TYPES

    if "timestamp" in contents:
        checkformat_utc_isoformat(contents["timestamp"])
    if "version" in contents:  # optional field for non-root metadata
        checkformat_natural_int(contents["version"])
    # TODO ✅: Ensure that expiration > timestamp, to help people not shoot
    #          themselves in the foot.


def checkformat_any_signature(signature: Any) -> Signature | GPGSignature:
    if not is_signature(signature) and not is_gpg_signature(signature):
        raise ValueError(
            "Expected either a hex string representing a raw ed25519 "
            "signature (see checkformat_signature) or a dictionary "
            "representing an OpenPGP/GPG signature "
            "(see checkformat_gpg_signature)."
        )

    return signature


def keyfiles_to_bytes(name):
    """
    Toy function.  Import an ed25519 key pair, in the forms of raw public and
    raw private keys, from name.pub and name.pri respectively.

    Cavalier about private key bytes.
    Does not perform input validation ('/'...).

    Return the 32 bytes of the private key object and the 32 bytes of the
    public key object, in that order.
    """
    with open(name + ".pri", "rb") as fobj:
        private_bytes = fobj.read()

    with open(name + ".pub", "rb") as fobj:
        public_bytes = fobj.read()

    return private_bytes, public_bytes


def keyfiles_to_keys(name):
    """
    Doesn't perform input validation.
    Import an ed25519 key pair, in the forms of raw public key
    bytes and raw private key bytes, from name.pub and name.pri respectively.
    Cavalier about private key bytes.
    Return a private key object and public key object, in that order.
    """
    private_bytes, public_bytes = keyfiles_to_bytes(name)

    private = PrivateKey.from_bytes(private_bytes)
    public = PublicKey.from_bytes(public_bytes)

    return private, public


def checkformat_key(key: Any) -> ed25519.Ed25519PublicKey | ed25519.Ed25519PrivateKey:
    """
    Enforces expectation that argument is an object of type
    cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PublicKey or
    cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PrivateKey.
    """
    if not isinstance(key, (ed25519.Ed25519PublicKey, ed25519.Ed25519PrivateKey)):
        raise TypeError(
            "Expected an Ed25519PublicKey or Ed25519PrivateKey object "
            'from the "cryptography" library.  Received object of type '
            + str(type(key))
            + " instead."
        )

    return key


def iso8601_time_plus_delta(delta):
    """
    Applies a datetime.timedelta to the current time in UTC with microseconds
    stripped, then converts to ISO8601 format and appends a 'Z' indicating that
    it is UTC time, not local time.  We only deal with UTC times!

    This is used for two purposes:
     - get current time in ISO8601 format, by passing in a 0 timedelta
     - get ISO8601 UTC timestamp for expiration dates

    regex for time: '^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}Z$'
    """
    checkformat_expiration_distance(delta)

    unix_expiry = datetime.utcnow().replace(microsecond=0) + delta

    return unix_expiry.isoformat() + "Z"



# Copyright (C) 2019 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
"""
This module contains functions that sign data using ed25519 keys, via the
pyca/cryptography library.  Functions that perform OpenPGP-compliant (e.g. GPG)
signing are provided instead in root_signing.

Function Manifest for this Module:
    serialize_and_sign
    wrap_as_signable
    sign_signable
"""
from binascii import hexlify
from copy import deepcopy

from .common import (
    SUPPORTED_SERIALIZABLE_TYPES,
    PrivateKey,
    PublicKey,
    canonserialize,
    checkformat_hex_key,
    checkformat_key,
    checkformat_signable,
    checkformat_signature,
    checkformat_string,
    load_metadata_from_file,
    write_metadata_to_file,
)


def serialize_and_sign(obj, private_key):
    """
    Given a JSON-compatible object, does the following:
     - serializes the dictionary as utf-8-encoded JSON, lazy-canonicalized
       such that any dictionary keys in any dictionaries inside <dictionary>
       are sorted and indentation is used and set to 2 spaces (using json lib)
     - creates a signature over that serialized result using private_key
     - returns that signature as a hex string

    See comments in common.canonserialize()

    Arguments:
      obj: a JSON-compatible object -- see common.canonserialize()
      private_key: a conda_content_trust.common.PrivateKey object

    # TODO ✅: Consider taking the private key data as a hex string instead?
    #          On the other hand, it's useful to support an object that could
    #          obscure the key (or provide an interface to a hardware key).
    """

    # Try converting to a JSON string.
    serialized = canonserialize(obj)

    signature_as_bytes = private_key.sign(serialized)

    signature_as_hexstr = hexlify(signature_as_bytes).decode("utf-8")

    return signature_as_hexstr


def wrap_as_signable(obj):
    """
    Given a JSON-serializable object (dictionary, list, string, numeric, etc.),
    returns a wrapped copy of that object:

        {'signatures': {},
         'signed': <deep copy of the given object>}

    Expects strict typing matches (not duck typing), for no good reason.
    (Trying JSON serialization repeatedly could be too time consuming.)

    TODO: ✅ Consider whether or not the copy can be shallow instead, for speed.

    Raises ❌TypeError if the given object is not a JSON-serializable type per
    SUPPORTED_SERIALIZABLE_TYPES
    """
    if not type(obj) in SUPPORTED_SERIALIZABLE_TYPES:
        raise TypeError(
            "wrap_dict_as_signable requires a JSON-serializable object, "
            "but the given argument is of type " + str(type(obj)) + ", "
            "which is not supported by the json library functions."
        )

    # TODO: ✅ Later on, consider switching back to TUF-style
    #          signatures-as-a-list.  (Is there some reason it's saner?)
    #          Going with my sense of what's best now, which is dicts instead.
    #          It's simpler and it naturally avoids duplicates.  We don't do it
    #          this way in TUF, but we also don't depend on it being an ordered
    #          list anyway, so a dictionary is probably better.

    return {"signatures": {}, "signed": deepcopy(obj)}


def sign_signable(signable, private_key):
    """
    Given a JSON-compatible signable dictionary (as produced by calling
    wrap_dict_as_signable with a JSON-compatible dictionary), calls
    serialize_and_sign on the enclosed dictionary at signable['signed'],
    producing a signature, and places the signature in
    signable['signatures'], in an entry indexed by the public key
    corresponding to the given private_key.

    Updates the given signable in place, returning nothing.
    Overwrites if there is already an existing signature by the given key.

    # TODO ✅: Take hex string keys for sign_signable and serialize_and_sign
    #          instead of constructed PrivateKey objects?  Add the comment
    #          below if so:
    # # Unlike with lower-level functions, both signatures and public keys are
    # # always written as hex strings.

    Raises ❌TypeError if the given object is not a JSON-serializable type per
    SUPPORTED_SERIALIZABLE_TYPES
    """
    # Argument checking
    checkformat_key(private_key)
    checkformat_signable(signable)
    # if not is_a_signable(signable):
    #     raise TypeError(
    #             'Expected a signable dictionary; the given argument of type ' +
    #             str(type(signable)) + ' failed the check.')

    # private_key = PrivateKey.from_hex(private_key_hex)

    signature_as_hexstr = serialize_and_sign(signable["signed"], private_key)

    public_key_as_hexstr = PublicKey.to_hex(private_key.public_key())

    # To fit a general format, we wrap it this way, instead of just using the
    # hexstring.  This is because OpenPGP signatures that we use for root
    # signatures look similar and have a few extra fields beyond the signature
    # value itself.
    signature_dict = {"signature": signature_as_hexstr}

    checkformat_signature(signature_dict)

    # TODO: ✅⚠️ Log a warning in whatever conda's style is (or conda-build):
    #
    # if public_key_as_hexstr in signable['signatures']:
    #   warn(    # replace: log, 'warnings' module, print statement, whatever
    #           'Overwriting existing signature by the same key on given '
    #           'signable.  Public key: ' + public_key + '.')

    # Add signature in-place, in the usual signature format.
    signable["signatures"][public_key_as_hexstr] = signature_dict


def sign_all_in_repodata(fname, private_key_hex):
    """
    Given a repodata.json filename, reads the "packages" entries in that file,
    and produces a signature over each artifact, with the given key.  The
    signatures are then placed in a "signatures" entry parallel to the
    "packages" entry in the json file.  The file is overwritten.

    Arguments:
        fname: filename of a repodata.json file
        private_key_hex:
            a private ed25519 key value represented as a 64-char hex string
    """
    checkformat_hex_key(private_key_hex)
    checkformat_string(fname)
    # TODO ✅⚠️: Consider filename validation.  What does conda use for that?

    private = PrivateKey.from_hex(private_key_hex)
    public_hex = PublicKey.to_hex(private.public_key())

    # Loading the whole file at once instead of reading it as we go, because
    # it's less complex and this only needs to run repository-side.
    repodata = load_metadata_from_file(fname)
    # with open(fname, 'rb') as fobj:
    #     repodata = json.load(fname)

    # TODO ✅: Consider more validation for the gross structure expected of
    #            repodata.json
    if "packages" not in repodata:
        raise ValueError('Expected a "packages" entry in given repodata file.')

    # Add an empty 'signatures' dict to repodata.
    # If it's already there for whatever reason, we replace it entirely.  This
    # avoids leaving existing signatures that might not get replaced -- e.g. if
    # the artifact is not in the "packages" dict, but is in the "signatures"
    # dict for some reason.  What comes out of this process will be limited to
    # what we sign in this function.
    repodata["signatures"] = {}

    for artifact_name, metadata in repodata["packages"].items():
        # TODO ✅: Further consider the significance of the artifact name
        #          itself not being part of the signed metadata.  The info used
        #          to generate the name (package name + version + build) is
        #          part of the signed metadata, but the full name is not.
        #          Keep in mind attacks that swap metadata among artifacts;
        #          signatures would still read as correct in that circumstance.
        signature_hex = serialize_and_sign(metadata, private)

        # To fit a general format, we wrap it this way, instead of just using
        # the hexstring.  This is because OpenPGP signatures that we use for
        # root signatures look similar and have a few extra fields beyond the
        # signature value itself.
        signature_dict = {"signature": signature_hex}

        checkformat_signature(signature_dict)

        repodata["signatures"][artifact_name] = {public_hex: signature_dict}

    # Repeat for the .conda packages in 'packages.conda'.
    for artifact_name, metadata in repodata.get("packages.conda", {}).items():
        signature_hex = serialize_and_sign(metadata, private)
        repodata["signatures"][artifact_name] = {
            public_hex: {"signature": signature_hex}
        }

    # Note: takes >0.5s on a macbook for large files
    write_metadata_to_file(repodata, fname)
